from typing import Dict, Optional, List, Type

from great_expectations.cli.datasource import DatasourceTypes
from kedro.io import AbstractDataSet


def identify_dataset_type(
    dataset: AbstractDataSet,
    pandas_datasets: Optional[List[Type[AbstractDataSet]]] = None,
    spark_datasets: Optional[List[Type[AbstractDataSet]]] = None,
) -> Optional[DatasourceTypes]:
    from kedro.extras.datasets.spark import SparkDataSet
    from kedro.extras.datasets.pandas import CSVDataSet, ExcelDataSet

    default_pandas_datasets = [
        CSVDataSet,
        ExcelDataSet,
    ]

    default_spark_datasets = [SparkDataSet]

    if pandas_datasets is None:
        pandas_datasets = []
    if spark_datasets is None:
        spark_datasets = []

    pandas_datasets += default_pandas_datasets
    spark_datasets += default_spark_datasets

    def _dataset_isinstance_of_list(
        target_dataset: AbstractDataSet, dataset_list: List[Type[AbstractDataSet]]
    ) -> bool:
        return any(
            [isinstance(target_dataset, dataset_type) for dataset_type in dataset_list]
        )

    if _dataset_isinstance_of_list(dataset, pandas_datasets):
        return DatasourceTypes.PANDAS
    elif _dataset_isinstance_of_list(dataset, spark_datasets):
        return DatasourceTypes.SPARK
    else:
        return None


def generate_datasource_name(dataset_name: str) -> str:
    return f"{dataset_name}___kedro_great_datasource"


def get_ge_class(datasource_type: DatasourceTypes) -> Optional[str]:
    return {
        DatasourceTypes.PANDAS: "PandasDataset",
        DatasourceTypes.SPARK: "SparkDFDataset",
    }.get(datasource_type)


def get_suite_name(
    expectations_map: Dict[str, str], dataset_name: str, suite_type: Optional[str]
) -> str:
    if suite_type is None:
        target_expectation_suite_name = (
            f"{expectations_map.get(dataset_name, dataset_name)}"
        )
    else:
        target_expectation_suite_name = (
            f"{expectations_map.get(dataset_name, dataset_name)}.{suite_type}"
        )
    return target_expectation_suite_name



import os
from pathlib import Path

import click
from great_expectations import DataContext
from great_expectations.cli import toolkit
from great_expectations.cli.cli_messages import (
    LETS_BEGIN_PROMPT,
    RUN_INIT_AGAIN,
    SETUP_SUCCESS,
)
from great_expectations.cli.init import _get_full_path_to_ge_dir
from great_expectations.cli.util import cli_message
from great_expectations.exceptions import DataContextError

from .datasource import generate_datasources
from .suite import generate_basic_suites


@click.command()
@click.option(
    "--target-directory",
    "-d",
    default="./",
    help="The root of the project directory where you want to initialize Great Expectations.",
)
@click.option(
    "--usage-stats/--no-usage-stats",
    help="By default, usage statistics are enabled unless you specify the --no-usage-stats flag.",
    default=True,
)
def init(target_directory, usage_stats):
    """
    Generate a new Great Expectations project configuration and
    fill in the Datasources and Suites based on the kedro catalog
    """
    from kedro.framework.context import load_context

    target_directory = os.path.abspath(target_directory)
    ge_dir = _get_full_path_to_ge_dir(target_directory)

    if not DataContext.does_config_exist_on_disk(ge_dir):
        if not click.confirm(LETS_BEGIN_PROMPT, default=True):
            cli_message(RUN_INIT_AGAIN)
            # TODO ensure this is covered by a test
            exit(0)
        try:
            DataContext.create(target_directory, usage_statistics_enabled=usage_stats)
            cli_message(SETUP_SUCCESS)
        except DataContextError as e:
            cli_message("<red>{}</red>".format(e.message))
            exit(5)

    if click.confirm("Generate Datasources based on Kedro Context?", default=True):
        kedro_context = load_context(Path.cwd())
        ge_context = toolkit.load_data_context_with_error_handling(ge_dir)
        new_datasources = generate_datasources(kedro_context, ge_context)
        if new_datasources:
            cli_message(
                "Added {} New datasources to your project.".format(len(new_datasources))
            )

    if click.confirm(
        "Generate Basic Validation Suites based on Kedro Context?", default=True
    ):
        kedro_context = load_context(Path.cwd())
        ge_context = toolkit.load_data_context_with_error_handling(ge_dir)
        new_datasources = generate_basic_suites(kedro_context, ge_context)
        if new_datasources:
            cli_message(
                "Added {} New datasources to your project.".format(len(new_datasources))
            )



import datetime
import logging
from typing import Any, Dict, List

import great_expectations as ge
from great_expectations.core.batch import Batch
from great_expectations.core.id_dict import BatchKwargs
from great_expectations.datasource.types import BatchMarkers
from great_expectations.validator.validator import Validator
from kedro.framework.hooks import hook_impl
from kedro.io import DataCatalog, AbstractDataSet

from ..data import (
    identify_dataset_type,
    get_suite_name,
    get_ge_class,
    generate_datasource_name,
)


class KedroGreat:
    DEFAULT_SUITE_TYPES = ["warning", "basic", None]

    def __init__(
        self,
        expectations_map: Dict = None,
        suite_types: List[str] = None,
        run_before_node: bool = True,
        run_after_node: bool = False,
    ):
        if expectations_map is None:
            expectations_map = {}
        if suite_types is None:
            suite_types = []
        suite_types += KedroGreat.DEFAULT_SUITE_TYPES
        self.expectations_map = expectations_map
        self.suite_types = suite_types

        self.expectation_context = ge.data_context.DataContext()
        self.expectation_suite_names = set(
            self.expectation_context.list_expectation_suite_names()
        )
        self._before_node_run = run_before_node
        self._after_node_run = run_after_node
        self.logger = logging.getLogger("KedroGreat")
        self._finished_suites = set()

    @hook_impl
    def before_node_run(
        self, catalog: DataCatalog, inputs: Dict[str, Any], run_id: str
    ) -> None:
        if self._before_node_run:
            self._run_validation(catalog, inputs, run_id)

    @hook_impl
    def after_node_run(
        self, catalog: DataCatalog, outputs: Dict[str, Any], run_id: str
    ) -> None:
        if self._after_node_run:
            self._run_validation(catalog, outputs, run_id)

    def _run_validation(self, catalog: DataCatalog, data: Dict[str, Any], run_id: str):
        for dataset_name, dataset_value in data.items():
            ran_suite_for_dataset = False
            for suite_type in self.suite_types:

                target_suite_name = get_suite_name(
                    self.expectations_map, dataset_name, suite_type
                )
                if (
                    target_suite_name not in self.expectation_suite_names
                    or target_suite_name in self._finished_suites
                ):
                    continue

                dataset = catalog._get_dataset(dataset_name)
                dataset_type = identify_dataset_type(dataset)
                dataset_class = get_ge_class(dataset_type)
                if dataset_class is None:
                    self.logger.warning(
                        f"Unsupported DataSet Type: {dataset_name}({type(dataset)})"
                    )
                    continue

                self._run_suite(dataset_name, dataset, target_suite_name, run_id)
                self._finished_suites.add(target_suite_name)
                ran_suite_for_dataset = True

            if not ran_suite_for_dataset:
                self.logger.warning(
                    f"Missing Expectation Suite for DataSet: {dataset_name}"
                )

    def _run_suite(
        self,
        dataset_name: str,
        dataset: AbstractDataSet,
        target_expectation_suite_name: str,
        run_id: str,
    ):
        target_suite = self.expectation_context.get_expectation_suite(
            target_expectation_suite_name
        )
        dataset_path = str(dataset._filepath)
        df = dataset.load()
        batch_markers = (
            BatchMarkers(
                {
                    "ge_load_time": datetime.datetime.now(
                        datetime.timezone.utc
                    ).strftime("%Y%m%dT%H%M%S.%fZ")
                }
            ),
        )
        batch = Batch(
            "kedro",
            batch_kwargs=BatchKwargs(
                {
                    "path": dataset_path,
                    "datasource": generate_datasource_name(dataset_name),
                }
            ),
            data=df,
            batch_parameters=None,
            batch_markers=batch_markers,
            data_context=self.expectation_context,
        )
        v = Validator(
            batch=batch,
            expectation_suite=target_suite,
            expectation_engine={
                "module_name": "great_expectations.dataset",
                "class_name": identify_dataset_type(dataset),
            },
        )
        validator_dataset_batch = v.get_dataset()
        return self.expectation_context.run_validation_operator(
            "action_list_operator", [validator_dataset_batch], run_id=run_id
        )
